// In this version, the Porshe hierarchy is equipped with native behavior
// (paint and recycle) which requires modifying the whole hierarchy by hand.
// For simplification, we use only structs, don't bother with virtual methods
// etc.

#include <iostream>
#include <ostream>
#include <vector>

struct Wheel
{
  Wheel ()        { std::cout << "Created a wheel."   << std::endl; };

  void paint ()   { std::cout << "Painting a wheel."  << std::endl; };
  void recycle () { std::cout << "Recycling a wheel." << std::endl; };
};

struct Body
{
  Body ()
  {
    for (int i = 0; i < 4; i++) wheels_.push_back (Wheel ());
    std::cout << "Created a body." << std::endl;
  };

  void paint ()
  {
    std::cout << "Painting the body." << std::endl;
    for (std::vector<Wheel>::iterator i = wheels_.begin ();
	 i != wheels_.end ();
	 i++)
      (*i).paint ();
  };
  void recycle ()
  {
    std::cout << "Recycling the body." << std::endl;
    for (std::vector<Wheel>::iterator i = wheels_.begin ();
	 i != wheels_.end ();
	 i++)
      (*i).recycle ();
  };

  std::vector<Wheel> wheels_;
};

struct Engine
{
  Engine ()       { std::cout << "Created an engine."    << std::endl; };

  void paint ()   { std::cout << "Painting the engine."  << std::endl; };
  void recycle () { std::cout << "Recycling the engine." << std::endl; };
};

struct Porshe
{
  Porshe () { std::cout << "Created a Porshe." << std::endl; };

  void paint ()
  {
    std::cout << "Painting the porshe." << std::endl;
    engine_.paint ();
    body_.paint ();
  };
  void recycle ()
  {
    std::cout << "Recycling the porshe." << std::endl;
    engine_.recycle ();
    body_.recycle ();
  };

  Body body_;
  Engine engine_;
};


int main (int argc, char *argv[])
{
  Porshe porshe;

  porshe.paint ();
  porshe.recycle ();
}
