(defclass wheel () ())

(defclass body ()
  ((wheels :initform (loop :for i :from 1 :upto 4
			   :for wheel = (make-instance 'wheel)
			   :collect wheel)
	   :accessor wheels)))

(defclass engine () ())

(defclass porshe ()
  ((engine :initform (make-instance 'engine)
	   :accessor engine)
   (body :initform (make-instance 'body)
	 :accessor body)))


;; In fact, ACCEPT should just be named VISIT now ;-)
(defgeneric accept (object visitor-function)
  ;; #### NOTE: for simple (non aggregative) objects like WHEEL and ENGINE,
  ;; the ACCEPT function will just call VISITOR-FUNCTION on OBJECT. This can
  ;; be unified in a (non-specialized) default method.
  (:method (object visitor-function)
    (funcall visitor-function object))
  (:method ((body body) visitor-function)
    (funcall visitor-function body)
    (dolist (wheel (wheels body))
      (accept wheel visitor-function)))
  (:method ((porshe porshe) visitor-function)
    (funcall visitor-function porshe)
    (accept (engine porshe) visitor-function)
    (accept (body porshe) visitor-function)))

;; No more need for PAINT-PORSHE-VISITOR or RECYCLE-PORSHE-VISITOR. Instead,
;; we just need to define two generic (visiting) functions. The failure to
;; implement one of the methods below would trigger a run-time error.
(defgeneric paint (object)
  (:method ((wheel wheel))
    (format t "Painting a wheel.~%"))
  (:method ((body body))
    (format t "Painting the body.~%"))
  (:method ((engine engine))
    (format t "Painting the engine.~%"))
  (:method ((porshe porshe))
    (format t "Painting the Porshe.~%")))

(defgeneric recycle (object)
  (:method ((wheel wheel))
    (format t "Recycling a wheel.~%"))
  (:method ((body body))
    (format t "Recycling the body.~%"))
  (:method ((engine engine))
    (format t "Recycling the engine.~%"))
  (:method ((porshe porshe))
    (format t "Recycling the Porshe.~%")))


(defun main ()
  (let ((porshe (make-instance 'porshe)))
    (accept porshe #'paint)
    (accept porshe #'recycle))
  (quit))
