;; Hmmm, but I hate to have to write down the MAP-OBJECT methods by hand every
;; time, knowing that I really want to map over all objects slots. So I can
;; use the MOP to do that automatically. The function below implements
;; automatic mapping for all objects slots, plus for lists, which is needed
;; for the engine wheels.
(defgeneric map-object (func object)
  ;; All standard objects are mappable here. You can prevent this by creating
  ;; a "mappable" class, have the desired classes inherit from it, and
  ;; specialize the method below on mappable objects only. You then have to
  ;; provide a default method that does nothing for the others.
  (:method (func (object standard-object))
    (funcall func  object)
    (mapc
	(lambda (slot)
	  (map-object func
		      ;; #### NOTE: SB-MOP is for SBCL.
		      (slot-value object (sb-mop:slot-definition-name slot))))
      (sb-mop:class-direct-slots (class-of object))))
  (:method (func (object list)) (mapc func object)))


(defclass wheel () ())

(defclass body ()
  ((wheels :initform (loop :for i :from 1 :upto 4
			   :for wheel = (make-instance 'wheel)
			   :collect wheel)
	   :accessor wheels)))

;; No more need for: (defmethod map-object (func (body body)) ...)

(defclass engine () ())

(defclass porshe ()
  ((engine :initform (make-instance 'engine)
	   :accessor engine)
   (body :initform (make-instance 'body)
	 :accessor body)))

;; No more need for: (defmethod map-object (func (porshe porshe)) ...)

(defgeneric paint (object)
  (:method ((wheel wheel))
    (format t "Painting a wheel.~%"))
  (:method ((body body))
    (format t "Painting the body.~%"))
  (:method ((engine engine))
    (format t "Painting the engine.~%"))
  (:method ((porshe porshe))
    (format t "Painting the Porshe.~%")))

(defgeneric recycle (object)
  (:method ((wheel wheel))
    (format t "Recycling a wheel.~%"))
  (:method ((body body))
    (format t "Recycling the body.~%"))
  (:method ((engine engine))
    (format t "Recycling the engine.~%"))
  (:method ((porshe porshe))
    (format t "Recycling the Porshe.~%")))

(defun main ()
  (let ((porshe (make-instance 'porshe)))
    (map-object #'paint porshe)
    (map-object #'recycle porshe))
  (quit))
