;; Now, suppose I want to track the object nesting level. Normally, this would
;; require having a specific visiting mechanism for alternatively incrementing
;; and decrementing the level each and every time we start/finish visiting an
;; object. But CLOS already has an infrastructure that let us do so very
;; simply.

(defgeneric map-object (func object)
  (:method (func (object standard-object))
    (funcall func  object)
    (mapc
	(lambda (slot)
	  (funcall #'map-object func
		   ;; That's where something should normally be done about the
		   ;; current nesting level.
		   ;; #### NOTE: SB-MOP is for SBCL.
		   (slot-value object (sb-mop:slot-definition-name slot))))
      (sb-mop:class-direct-slots (class-of object))))
  (:method (func (object list))
    (mapc func object)))


(defclass wheel () ())

(defclass body ()
  ((wheels :initform (loop :for i :from 1 :upto 4
			   :for wheel = (make-instance 'wheel)
			   :collect wheel)
	   :accessor wheels)))

(defclass engine () ())

(defclass porshe ()
  ((engine :initform (make-instance 'engine)
	   :accessor engine)
   (body :initform (make-instance 'body)
	 :accessor body)))

;; Instead of dynamically adding/removing before/after methods, we can also
;; use the fact that CLOS has multi-methods to specialize MAP-OBJECT on the
;; first argument (as well as the second), and EQL-specialize it on the
;; function below, so that those methods are only applicable then.
(let ((depth 0))
  (defun print-depth (obj)
    (format t "~S, current level: ~A~%" (class-name (class-of obj)) depth))
  (defmethod map-object :before ((func (eql #'print-depth)) object)
    (incf depth))
  (defmethod map-object :after ((func (eql #'print-depth)) object)
    (decf depth)))

(defun main ()
  (let* ((porshe (make-instance 'porshe)))
    (map-object #'print-depth porshe))
  (values))
